unit datalogtrack;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Buttons, ExtCtrls, ColorBox, Synaser;

const
  ezredids	: array[1..24] of string = (
    'X1',
    'X2',
    'X3',
    'X4',
    'X5',
    'X6',
    'X7',
    'X8',
    'FX1',
    'FX2',
    'AIN1',
    'AIN2',
    'Y1',
    'Y2',
    'Y3',
    'Y4',
    'Y5',
    'Y6',
    'Y7',
    'Y8',
    'AOUT1',
    'AOUT2',
    'ENCODER',  // high-1
    'SPEED'		// high
    );


type
  TValAry		= array[1..MAXLONGINT div 32] of single;
  TTrackData	= record
    name		: string[24];
    comnum		: integer;		// serial port number: 0=free slot
    comport		: TBlockSerial;
    datid		: integer;		// ezred value: 1..8=x1-8 ...see array above
    acqscale	: single;		// acquisition scaling
    acqoffs		: single;		// acquisition offset
    oldencod	: integer;		// old encoder for speed calculation
    scrzom		: single;		// screen zoom
    scrofs		: integer;		// screen offset
    scrcolor	: TColor;
    // output conditioning
    condONsam	: integer;		// time to go on (0=never)
    condOFsam	: integer;		// time to stay on (0=sudden off)
    condcnt		: integer;
    condison	: boolean;
    data		: ^TValAry;
  end;

  { TfmTrack }
  TfmTrack = class(TForm)
    btCancel: TBitBtn;
    btDelete: TBitBtn;
    btClose: TBitBtn;
    cbPortaser: TComboBox;
    cbDatum: TComboBox;
    cbColor: TColorBox;
    edEvery: TEdit;
    edLong: TEdit;
    edScale: TEdit;
    edOffset: TEdit;
    edName: TEdit;
    gbOuts: TGroupBox;
    Label1: TLabel;
    Label10: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    lPollRes: TLabel;
    lRData: TLabel;
    lComport: TLabel;
    tmPoll: TTimer;

    procedure btCancelClick(Sender: TObject);
    procedure btCloseClick(Sender: TObject);
    procedure btDeleteClick(Sender: TObject);
    procedure cbDatumSelect(Sender: TObject);
    procedure cbPortaserDropDown(Sender: TObject);
    procedure cbPortaserSelect(Sender: TObject);
    procedure edEveryExit(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure TestPort;
    procedure Prepare;
    procedure tmPollTimer(Sender: TObject);
  private
    { private declarations }
  public
    { public declarations }
  end;

var
  fmTrack	: TfmTrack;
  tracks	: array[1..16] of TTrackData;
  curez		: integer;
  TestPt	: TBlockSerial	= NIL;
  lastwok	: integer = 0;			// last working port
  lastezst	: string;				// last reply from ez poll

  samples	: integer = 0;			// number of samples present
  datamodi	: boolean = false;
  arylen	: integer = 0;			// array length
  timebase	: integer = 0;
  portsopen	: boolean = false;


function  CreateEzPort(portnum: integer): TBlockSerial;
procedure OpenPorts;
procedure ClosePorts;

function  sendtoez(port: TBlockSerial; what: string): boolean;	// true if error
function  askez(port: TBlockSerial; datid,ezid: integer; sc,os: real): real;
function  commandez(port: TBlockSerial; command: string): boolean;	// true if error

function  DoSaveFiles(fname: string): integer;
function  DoLoadFiles(fname: string): integer;

implementation
Uses Registry, LCLType, IniFiles;
{$R *.lfm}

procedure OpenPorts;
var
  i,k		: integer;

begin
  portsopen := false;
  for i := 1 to High(tracks) do with tracks[i] do begin
    // se di quelle precedenti ce n'è una, usare quella
    if comnum<1 then continue;
    for k := 1 to i-1 do begin
      if tracks[k].comnum=comnum then begin
        comport := tracks[k].comport;
        break
      end;
    end;
    if comport=NIL
      then comport := CreateEzPort(comnum);
    if comport<>NIL then portsopen := true;
  end
end;

procedure ClosePorts;
var
  i,k		: integer;

begin
  portsopen := false;
  for i := 1 to high(tracks) do with tracks[i] do begin
	if comnum=0 then continue;
    if comport<>NIL then begin
      for k := i+1 to high(tracks)
        do if tracks[k].comport=comport then tracks[k].comport := NIL;
      comport.free;
    end;
    comport := NIL;
  end
end;

procedure TfmTrack.Prepare;
var
  i,k		: integer;

begin
  cbPortaserDropDown(NIL);
  Caption := 'Track properties: '+tracks[curez].name;

  cbDatum.Clear;
  cbDatum.Items.add(' (None) ');
  for i := 1 to high(ezredids) do cbDatum.Items.add(ezredids[i]);
  cbDatum.ItemIndex := 0;

  edName.text := tracks[curez].name;
  cbDatum.ItemIndex := tracks[curez].datid;
  if tracks[curez].comnum=0
    then tracks[curez].comnum := lastwok;
  cbPortaser.ItemIndex := cbPortaser.Items.IndexOf('COM'+inttostr(tracks[curez].comnum));
  if cbPortaser.ItemIndex=-1 then begin
    // cbPortaser.Text := 'Select...';
    lComport.Caption := '<- Select a port';
  end;
  if tracks[curez].acqscale=0 then tracks[curez].acqscale := 1;
  edScale.text := FloatToStr(tracks[curez].acqscale);
  edOffset.text := FloatToStr(tracks[curez].acqoffs);

  if tracks[curez].scrcolor=clBlack then tracks[curez].scrcolor := clYellow;
  cbColor.Selected := tracks[curez].scrcolor;

  edEvery.text := inttostr(tracks[curez].condONsam);
  edLong.text  := inttostr(tracks[curez].condOFsam);
  edEveryExit(edEvery);
  edEveryExit(edLong);
  cbDatumSelect(NIL);

  ClosePorts;

  tmPoll.Enabled := true;

  TestPort;
end;

function sendtoez(port: TBlockSerial; what: string): boolean;	// true if error
var
  tries		: integer;

begin
  if port=NIL then begin
    result := true;
    exit
  end;
  result := false;

  what := what+#13;
  for tries := 1 to 4 do begin
    port.Purge;
    port.SendString(what);
    lastezst := port.Recvstring(10);
    if (port.LastError=0) and (lastezst<>'') then exit;
    sleep(8);
  end;
  result := true;	// error
end;

function commandez(port: TBlockSerial; command: string): boolean;	// true if error
begin
  result := true;
  if sendtoez(port, '>'+command) then exit;
  if lastezst='OK' then result := false;
end;

function askez(port: TBlockSerial; datid,ezid: integer; sc,os: real): real;
var
  tries		: integer;
  res		: integer;
  res2		: integer;

  function poll(what: string; docheck: boolean=true): boolean;		// true if ok
  begin
    result := not sendtoez(port, '?'+what);
    if result and docheck then begin
      if copy(lastezst,1,length(what))<>what then result := false
    end
  end;

begin
  result := 0;
  if port=NIL then exit;
  if datid<1 then exit;
  for tries := 1 to 4 do begin
    if datid<=22 then begin
      if poll(ezredids[datid]) then begin
        res := strtointdef(copy(lastezst, 2+length(ezredids[datid]), 12), 0);
		result := res*sc + os;
        exit
      end
    end else if datid=23 then begin
      // encoder
      if poll('ENCODER', false) then if copy(lastezst,1,4)='ENC=' then begin
        res := strtointdef(copy(lastezst, 5, 12), 0);
		result := res*sc + os;
        exit
      end
    end else if datid=24 then begin
      // speed
      if ezid in [1..high(tracks)] then begin
        if poll('ENCODER', false) then if copy(lastezst,1,4)='ENC=' then begin
          res := strtointdef(copy(lastezst, 5, 12), 0);
          res2 := res-tracks[ezid].oldencod;
          if res2<0 then inc(res2,65535);
          lastezst := 'SPEED='+inttostr(res2);
          result := res2*sc + os;
		  tracks[ezid].oldencod := res;
          exit
        end
      end
    end
  end;
  lastezst := 'Timeout for '+ezredids[datid];
end;

procedure TfmTrack.tmPollTimer(Sender: TObject);
var
  st	: string;
  res	: real;

begin
  if TestPt=NIL then begin
    lRData.Caption := '- no port selected -';
    exit;
  end;
  if cbDatum.ItemIndex=0 then begin
    lRData.Caption := '<- no data to poll';
    exit
  end;
  if cbDatum.ItemIndex<=high(ezredids) then begin
    res := askez(testPt, cbDatum.ItemIndex, curez, StrToFloatDef(edScale.text, 1), StrToFloatDef(edOffset.Text, 0));
    lRData.Caption := lastezst;
    lPollRes.Caption := format('result=%5.2f', [res]);
  end
end;

procedure TfmTrack.cbPortaserDropDown(Sender: TObject);
var
  i			: integer;
  reg		: TRegistry;
  comports	: TStringList;
  names		: TStringList;

begin
  reg := Tregistry.Create;
  reg.RootKey := HKEY_LOCAL_MACHINE;
  reg.OpenKeyReadOnly('\hardware\devicemap\serialcomm');
  comports := TStringList.Create;
  names := TStringList.Create;
  reg.GetValueNames(names);
  cbPortaser.Clear;
  for i := 0 to names.Count-1 do cbPortaser.Items.Add(reg.ReadString(names[i]));

  comports.Free;
  names.Free;
  reg.free
end;

function  CreateEzPort(portnum: integer): TBlockSerial;
var
  st	: string;
  pt	: TBlockSerial;

begin
  result := NIL;
  if portnum=0 then exit;
  pt := TBlockSerial.Create;
  pt.Connect('COM'+inttostr(portnum));
  if pt.LastError<>0 then begin
    fmTrack.lComport.Caption := pt.LastErrorDesc;
    pt.Free;
    exit
  end;
  pt.Config(38400, 8, 'N', SB1, false,false);
  pt.ConvertLineEnd := true;

  pt.Flush;
  pt.Recvstring(20);
  pt.Purge;
  pt.SendString('?DEVICE'#13);
  st := pt.Recvstring(30);
  if st='' then begin
	fmTrack.lComport.Caption := 'No reply from this port';
    pt.Free;
    exit
  end;
  if pos('EZ-RED 1.', st)=0 then begin
	fmTrack.lComport.Caption := 'No EZ-Red connected here.';
    pt.Free;
    exit
  end;
  fmTrack.lComport.Caption := st;
  lastwok := portnum;

  result := pt;
end;

procedure TfmTrack.TestPort;
begin
  if TestPt<>NIL then TestPt.Free;
  TestPt := CreateEzPort( strtointdef( copy(cbPortaser.Text,4,3), 0) );
  if TestPt=NIL then exit;
end;

procedure TfmTrack.cbPortaserSelect(Sender: TObject);
var
  n		: integer;

begin
  n := strtoint( copy(cbPortaser.Text,4,3) );
  TestPort;
end;

procedure TfmTrack.FormShow(Sender: TObject);
begin
  edName.SetFocus;
end;

procedure TfmTrack.btDeleteClick(Sender: TObject);
begin
  if Application.MessageBox('Really delete track?', 'Xdatalogger', MB_YESNO)<>IDYES then exit;
  if TestPt<>NIL then TestPt.Free;
  TestPt := NIL;
  tmPoll.Enabled := false;
  ModalResult := mrIgnore;
end;

procedure TfmTrack.cbDatumSelect(Sender: TObject);
begin
  gbOuts.Visible := (cbDatum.ItemIndex>=13) and (cbDatum.ItemIndex<=13+10-1)
end;

function parseminsec(st: string): integer;
var
  i		: integer;

begin
  i := pos('s', st);
  if i=0 then i := pos('S', st);
  if i=0 then i := pos('m', st);
  if i=0 then i := pos('M', st);
  if i=0 then begin
    result := StrToIntDef(st, 0);
  end else begin
    result := StrToIntDef(copy(st,1,i-1), 0);
    if upcase(st[i])='S'
      then result := result*1000
      else result := result*1000*60
  end
end;

procedure TfmTrack.edEveryExit(Sender: TObject);
var
  i		: integer;

begin
  i := parseminsec(TEdit(sender).text);
  if (i>=60000) and (i mod 60000=0)
    then TEdit(Sender).text := inttostr(i div 60000)+'m'
    else if (i>=1000) and (i mod 1000=0)
      then TEdit(Sender).text := inttostr(i div 1000)+'s'
end;

procedure TfmTrack.btCloseClick(Sender: TObject);
var
  i		: integer;

begin
  if TestPt<>NIL then TestPt.Free;
  TestPt := NIL;
  tmPoll.Enabled := false;

  with tracks[curez] do begin
    name := edName.Text;
    tracks[curez].comnum := strtointdef( copy(cbPortaser.Text,4,3), 0);

    if comnum=0 then comnum := -1;
	datid := cbDatum.ItemIndex;
	acqscale := StrToFloatDef(edScale.Text, 1);
  	acqoffs  := StrToFloatDef(edOffset.Text, 0);
    scrcolor := cbColor.Selected;

    // output cond
    if gbOuts.visible then begin
      condONsam := parseminsec(edEvery.text);
      condOFsam := parseminsec(edLong.text);
	end else begin
      condONsam := 0;
      condOFsam := 0;
    end;
  end;
  datamodi := true;
end;

procedure TfmTrack.btCancelClick(Sender: TObject);
begin
  if TestPt<>NIL then TestPt.Free;
  TestPt := NIL;
end;

function  DoSaveFiles(fname: string): integer;
var
  i			: integer;
  tr		: integer;
  trname	: string;

begin
  if FileExists(fname) then begin
    if FileExists(fname+'.bak') then DeleteFile(fname+'.bak');
    RenameFile(fname, fname+'.bak');
  end;
  with TIniFile.Create(fname) do begin
    tr := 0;
    for i := 1 to high(tracks) do with tracks[i] do begin
      if comnum=0 then continue;
      inc(tr);
      trname := 'Track'+inttostr(tr);
      WriteString(trname, 'name', name);
      WriteInteger(trname, 'com', comnum);
      WriteInteger(trname, 'datid', datid);
      WriteFloat(trname, 'acqscale', acqscale);
      WriteFloat(trname, 'acqoffs', acqoffs);
      WriteFloat(trname, 'scrzom', scrzom);
      WriteInteger(trname, 'scrofs', scrofs);
      WriteInteger(trname, 'color', scrcolor);
      WriteInteger(trname, 'cycleon', condONsam);
      WriteInteger(trname, 'pulselen', condOFsam);
    end;
    // WriteInteger('Data','Samples', samples);
    WriteInteger('Data','Period', timebase);
    free
  end;
  result := 0
end;

function  DoLoadFiles(fname: string): integer;
var
  i			: integer;
  tr		: integer;
  trname	: string;

begin
  with TIniFile.Create(fname) do begin
    for i := 1 to high(tracks) do with tracks[i] do begin
      trname := 'Track'+inttostr(i);
      comnum := ReadInteger(trname, 'com', 0);
      if comnum=0 then continue;

      name := ReadString(trname, 'name', 'no name');
      datid := ReadInteger(trname, 'datid', 0);
      acqscale := ReadFloat(trname, 'acqscale', 1);
      acqoffs := ReadFloat(trname, 'acqoffs', 0);
      scrzom := ReadFloat(trname, 'scrzom', 4);
      scrofs := ReadInteger(trname, 'scrofs', 100);
      scrcolor  := TColor(ReadInteger(trname, 'color', clYellow));

      condONsam	:= ReadInteger(trname, 'cycleon', 0);
      condOFsam	:= ReadInteger(trname, 'pulselen', 0);

    end;
    // samples := ReadInteger('Data', 'Samples', 0);
	timebase := ReadInteger('Data','Period', 200);
    free
  end;
  result := 0
end;

end.

